@@ -23,13 +23,10 @@ ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.
 */
 
-#if (defined(__linux__) || defined(__NetBSD__) || defined(__FreeBSD__)) && \
-	defined(__i386__)
-
 #include <stdio.h>
 #include <string.h>
 
-#if defined(__linux__)
+#if defined(__linux__) && defined(__i386__)
 
 #include <asm/vm86.h>
 
@@ -54,152 +51,8 @@ OTHER DEALINGS IN THE SOFTWARE.
 #include <unistd.h>
 #include <fcntl.h>
 
-#include "lrmi.h"
-
-#define REAL_MEM_BASE 	((void *)0x10000)
-#define REAL_MEM_SIZE 	0x40000
-#define REAL_MEM_BLOCKS 	0x100
-
-struct mem_block {
-	unsigned int size : 20;
-	unsigned int free : 1;
-};
-
-static struct {
-	int ready;
-	int count;
-	struct mem_block blocks[REAL_MEM_BLOCKS];
-} mem_info = { 0 };
-
-static int
-real_mem_init(void)
-{
-	void *m;
-	int fd_zero;
-
-	if (mem_info.ready)
-		return 1;
-
-	fd_zero = open("/dev/zero", O_RDONLY);
-	if (fd_zero == -1) {
-		perror("open /dev/zero");
-		return 0;
-	}
-
-	m = mmap((void *)REAL_MEM_BASE, REAL_MEM_SIZE,
-	 PROT_READ | PROT_WRITE | PROT_EXEC,
-	 MAP_FIXED | MAP_PRIVATE, fd_zero, 0);
-
-	if (m == (void *)-1) {
-		perror("mmap /dev/zero");
-		close(fd_zero);
-		return 0;
-	}
-
-	close(fd_zero);
-
-	mem_info.ready = 1;
-	mem_info.count = 1;
-	mem_info.blocks[0].size = REAL_MEM_SIZE;
-	mem_info.blocks[0].free = 1;
-
-	return 1;
-}
-
-static void
-real_mem_deinit(void)
-{
-	if (mem_info.ready) {
-		munmap((void *)REAL_MEM_BASE, REAL_MEM_SIZE);
-		mem_info.ready = 0;
-	}
-}
-
-
-static void
-insert_block(int i)
-{
-	memmove(
-	 mem_info.blocks + i + 1,
-	 mem_info.blocks + i,
-	 (mem_info.count - i) * sizeof(struct mem_block));
-
-	mem_info.count++;
-}
-
-static void
-delete_block(int i)
-{
-	mem_info.count--;
-
-	memmove(
-	 mem_info.blocks + i,
-	 mem_info.blocks + i + 1,
-	 (mem_info.count - i) * sizeof(struct mem_block));
-}
-
-void *
-LRMI_alloc_real(int size)
-{
-	int i;
-	char *r = (char *)REAL_MEM_BASE;
-
-	if (!mem_info.ready)
-		return NULL;
-
-	if (mem_info.count == REAL_MEM_BLOCKS)
-		return NULL;
-
-	size = (size + 15) & ~15;
-
-	for (i = 0; i < mem_info.count; i++) {
-		if (mem_info.blocks[i].free && size < mem_info.blocks[i].size) {
-			insert_block(i);
-
-			mem_info.blocks[i].size = size;
-			mem_info.blocks[i].free = 0;
-			mem_info.blocks[i + 1].size -= size;
-
-			return (void *)r;
-		}
-
-		r += mem_info.blocks[i].size;
-	}
-
-	return NULL;
-}
-
-
-void
-LRMI_free_real(void *m)
-{
-	int i;
-	char *r = (char *)REAL_MEM_BASE;
-
-	if (!mem_info.ready)
-		return;
-
-	i = 0;
-	while (m != (void *)r) {
-		r += mem_info.blocks[i].size;
-		i++;
-		if (i == mem_info.count)
-			return;
-	}
-
-	mem_info.blocks[i].free = 1;
-
-	if (i + 1 < mem_info.count && mem_info.blocks[i + 1].free) {
-		mem_info.blocks[i].size += mem_info.blocks[i + 1].size;
-		delete_block(i + 1);
-	}
-
-	if (i - 1 >= 0 && mem_info.blocks[i - 1].free) {
-		mem_info.blocks[i - 1].size += mem_info.blocks[i].size;
-		delete_block(i);
-	}
-}
-
+#include "include/lrmi.h"
+#include "x86-common.h"
 
 #if defined(__linux__)
 #define DEFAULT_VM86_FLAGS 	(IF_MASK | IOPL_MASK)
@@ -279,52 +132,13 @@ int
 LRMI_init(void)
 {
 	void *m;
-	int fd_mem;
 
 	if (context.ready)
 		return 1;
 
-	if (!real_mem_init())
+	if (!LRMI_common_init())
 		return 0;
 
-	/*
-	 Map the Interrupt Vectors (0x0 - 0x400) + BIOS data (0x400 - 0x502)
-	 and the ROM (0xa0000 - 0x100000)
-	*/
-	fd_mem = open("/dev/mem", O_RDWR);
-
-	if (fd_mem == -1) {
-		real_mem_deinit();
-		perror("open /dev/mem");
-		return 0;
-	}
-
-	m = mmap((void *)0, 0x502,
-	 PROT_READ | PROT_WRITE | PROT_EXEC,
-	 MAP_FIXED | MAP_PRIVATE, fd_mem, 0);
-
-	if (m == (void *)-1) {
-		close(fd_mem);
-		real_mem_deinit();
-		perror("mmap /dev/mem");
-		return 0;
-	}
-
-	m = mmap((void *)0xa0000, 0x100000 - 0xa0000,
-	 PROT_READ | PROT_WRITE,
-	 MAP_FIXED | MAP_SHARED, fd_mem, 0xa0000);
-
-	if (m == (void *)-1) {
-		munmap((void *)0, 0x502);
-		close(fd_mem);
-		real_mem_deinit();
-		perror("mmap /dev/mem");
-		return 0;
-	}
-
-	close(fd_mem);
-
-
 	/*
 	 Allocate a stack
 	*/
@@ -783,6 +597,13 @@ run_vm86(void)
 			if (v == RETURN_TO_32_INT)
 				return 1;
 
+			fprintf(stderr, "Calling INT 0x%X (%04X:%04X)\n",
+					v,
+					get_int_seg(v),
+					get_int_off(v));
+			fprintf(stderr, " EAX is 0x%lX\n",
+					CONTEXT_REGS.REG(eax));
+
 			pushw(CONTEXT_REGS.REG(eflags));
 			pushw(CONTEXT_REGS.REG(cs));
 			pushw(CONTEXT_REGS.REG(eip));
@@ -1027,4 +848,8 @@ LRMI_int(int i, struct LRMI_regs *r)
 	return vret;
 }
 
-#endif /* (__linux__ || __NetBSD__ || __FreeBSD__) && __i386__ */
+size_t
+LRMI_base_addr(void)
+{
+	return 0;
+}
