@@ -21,7 +21,7 @@ version 2
 #include <sys/stat.h>
 #include <errno.h>
 
-#include "lrmi.h"
+#include "include/lrmi.h"
 #include "vbetool.h"
 
 #define access_ptr_register(reg_frame,reg) (reg_frame -> reg)
@@ -38,8 +38,14 @@ static struct pci_access *pacc;
 
 int main(int argc, char *argv[])
 {
-	if (!LRMI_init())
+	/* Don't bother checking for privilege if they only want usage() */
+	if (argc < 2)
+		goto usage;
+
+	if (!LRMI_init()) {
+		fprintf(stderr, "Failed to initialise LRMI (Linux Real-Mode Interface).\n");
 		return 1;
+	}
 
 	ioperm(0, 1024, 1);
 	iopl(3);
@@ -107,10 +113,24 @@ int main(int argc, char *argv[])
 		} else {
 			return disable_vga();
 		}
+	} else if (!strcmp(argv[1], "vbefp")) {
+		if (!strcmp(argv[2], "id")) {
+			return do_get_panel_id(0);
+		} else if (!strcmp(argv[2], "panelsize")) {
+			return do_get_panel_id(1);
+		} else if (!strcmp(argv[2], "getbrightness")) {
+			return do_get_panel_brightness();
+		} else if (!strcmp(argv[2], "setbrightness")) {
+			return do_set_panel_brightness(atoi(argv[3]));
+		} else if (!strcmp(argv[2], "invert")) {
+			return do_invert_panel();
+		} else {
+			return 1;
+		}
 	} else {
 	      usage:
 		fprintf(stderr,
-			"%s: Usage %s [[vbestate save|restore]|[vbemode set|get]|[vgamode]|[dpms on|off|standby|suspend|reduced]|[post]|[vgastate on|off]]\n",
+			"%s: Usage %s [[vbestate save|restore]|[vbemode set|get]|[vgamode]|[dpms on|off|standby|suspend|reduced]|[post]|[vgastate on|off]|[vbefp panelid|panelsize|getbrightness|setbrightness|invert]]\n",
 			argv[0], argv[0]);
 		return 1;
 	}
@@ -123,6 +143,7 @@ int do_vbe_service(unsigned int AX, unsigned int BX, reg_frame * regs)
 	const unsigned interrupt = 0x10;
 	unsigned function_sup;
 	unsigned success;
+	regs->ds = 0x0040;
 
 	access_ptr_register(regs, eax) = AX;
 	access_ptr_register(regs, ebx) = BX;
@@ -138,10 +159,15 @@ int do_vbe_service(unsigned int AX, unsigned int BX, reg_frame * regs)
 	function_sup = ((AX & 0xff) == 0x4f);
 	success = ((AX & 0xff00) == 0);
 
-	if (!success)
+	if (!success) {
+		fprintf(stderr, "Real mode call failed\n");
 		return -2;
-	if (!function_sup)
+	}
+
+	if (!function_sup) {
+		fprintf(stderr, "Function not supported\n");
 		return -3;
+	}
 
 	return access_ptr_register(regs, ebx);
 }
@@ -163,9 +189,7 @@ int do_real_post(unsigned pci_device)
 
 	/* This is all heavily cargo culted but seems to work */
 	r.edx = 0x80;
-	r.ss = 0x0030;
 	r.ds = 0x0040;
-	r.sp = 0xfffe;
 
 	if (!LRMI_call(&r)) {
 		fprintf(stderr,
@@ -204,7 +228,7 @@ void restore_state()
 {
 	struct LRMI_regs r;
 
-	void *data = NULL;
+	char *data = NULL;
 	char tmpbuffer[524288];
 	int i, length = 0;
 
@@ -231,8 +255,9 @@ void restore_state()
 	r.eax = 0x4f04;
 	r.ecx = 0xf;		/* all states */
 	r.edx = 2;		/* restore state */
-	r.es = (unsigned int) data >> 4;
-	r.ebx = (unsigned int) data & 0xf;
+	r.es = (unsigned int) (data - LRMI_base_addr()) >> 4;
+	r.ebx = (unsigned int) (data - LRMI_base_addr()) & 0xf;
+	r.ds = 0x0040;
 
 	if (!LRMI_int(0x10, &r)) {
 		fprintf(stderr,
@@ -249,7 +274,7 @@ void restore_state()
 void save_state()
 {
 	struct LRMI_regs r;
-	void *buffer;
+	char *buffer;
 	unsigned int size;
 
 	memset(&r, 0, sizeof(r));
@@ -257,6 +282,7 @@ void save_state()
 	r.eax = 0x4f04;
 	r.ecx = 0xf;		/* all states */
 	r.edx = 0;		/* get buffer size */
+	r.ds = 0x0040;
 
 	if (!LRMI_int(0x10, &r)) {
 		fprintf(stderr,
@@ -277,11 +303,18 @@ void save_state()
 
 	memset(&r, 0, sizeof(r));
 
+	fprintf(stderr, "Allocated buffer at %p (base is 0x%x)\n", buffer,
+			LRMI_base_addr());
+
 	r.eax = 0x4f04;
 	r.ecx = 0xf;		/* all states */
 	r.edx = 1;		/* save state */
-	r.es = (unsigned int) buffer >> 4;
-	r.ebx = (unsigned int) buffer & 0xf;
+	
+	r.es = (unsigned int) (buffer - LRMI_base_addr()) >> 4;
+	r.ebx = (unsigned int) (buffer - LRMI_base_addr()) & 0xf;
+	r.ds = 0x0040;
+
+	fprintf(stderr, "ES: 0x%04X EBX: 0x%04X\n", r.es, r.ebx);
 
 	if (!LRMI_int(0x10, &r)) {
 		fprintf(stderr, "Can't save video state (vm86 failure)\n");
@@ -326,6 +359,68 @@ int do_set_mode (int mode, int vga) {
 	return 0;
 }
 
+int do_get_panel_brightness() {
+	reg_frame regs;
+	int error;
+
+	memset(&regs, 0, sizeof(regs));
+
+	error = do_vbe_service(0x4f11, 0x05, &regs);
+
+	if (error<0) {
+		return error;
+	}
+
+	printf("%d\n",regs.ecx);
+
+	return 0;
+}
+
+int do_invert_panel() {
+	reg_frame regs;
+	int error;
+
+	memset(&regs, 0, sizeof(regs));
+
+	error = do_vbe_service(0x4f11, 0x02, &regs);
+
+	if (error<0) {
+		return error;
+	}
+
+	if ((regs.ebx & 0xff) == 0)
+		regs.ecx = 3;
+	else
+		regs.ecx = 0;
+
+	error = do_vbe_service(0x4f11, 0x0102, &regs);
+
+	if (error<0) {
+		return error;
+	}
+
+	return 0;
+}
+
+int do_set_panel_brightness(int brightness) {
+	reg_frame regs;
+	int error;
+
+	memset(&regs, 0, sizeof(regs));
+
+	regs.ecx = brightness;
+
+	error = do_vbe_service(0x4f11, 0x0105, &regs);
+
+	if (error<0) {
+		return error;
+	}
+
+	printf("%d\n",regs.ecx);
+
+	return 0;
+}
+
 int do_get_mode() {
 	reg_frame regs;
 	int error;
@@ -377,3 +472,65 @@ int disable_vga() {
 	outb(~0x01 & inb(0x102),  0x102);
 	return 0;
 }
+
+/* Based on xserver-xorg-driver-i810/src/i830_driver.c */
+struct panel_id {
+  int hsize:16, vsize:16;
+  int fptype:16;
+  int redbpp:8, greenbpp:8, bluebpp:8, reservedbpp:8;
+  int rsvdoffscrnmemsize:32, rsvdoffscrnmemptr:32;
+  char reserved[14];
+} __attribute__((packed));
+
+int do_get_panel_id(int just_dimensions)
+{
+  reg_frame r = {
+    .eax = 0x4f11,
+    .ebx = 0x0001
+  };
+  struct panel_id *id = LRMI_alloc_real(sizeof(struct panel_id));
+  r.es = (unsigned short)(((int)(id-LRMI_base_addr()) >> 4) & 0xffff);
+  r.edi = (unsigned long)(id-LRMI_base_addr()) & 0xf;
+
+  if(sizeof(struct panel_id) != 32)
+    return fprintf(stderr, "oops: panel_id, sizeof struct panel_id != 32, it's %d...\n", sizeof(struct panel_id)), 7;
+
+  if(real_mode_int(0x10, &r))
+    return fprintf(stderr, "Can't get panel id (vm86 failure)\n"), 8;
+
+  if((r.eax & 0xff) != 0x4f)
+    return fprintf(stderr, "Panel id function not supported\n"), 9;
+
+  if(r.eax & 0xff00)
+    {
+      if((r.eax & 0xff00) == 0x100)
+	fprintf(stderr, "Panel id read failed\n");
+      else
+	fprintf(stderr, "Panel id function not successful\n");
+      return 10;
+    }
+
+  if(!just_dimensions)
+    printf("size:\t%d %d\n"
+	   "type:\t%d\n"
+	   "bpp:\t%d %d %d %d\n",
+	   id->hsize, id->vsize,
+	   id->fptype,
+	   id->redbpp, id->greenbpp, id->bluebpp, id->reservedbpp);
+  else
+    printf("%dx%d\n", id->hsize, id->vsize);
+
+#if 0
+
+  /* Don't have a use for these and they don't seem to be documented.
+   * 320 appears to be 320kB of mapped memory and the following
+   * pointer is 0x1ffb8000 which is kernel mapping + 0xb8000 offset.
+   */
+  printf("ram:\t%dkB\n"
+	 "offset:\t%p\n",
+	 id->rsvdoffscrnmemsize,
+	 (void *)id->rsvdoffscrnmemptr);
+#endif
+
+  return 0;
+}
